/* $Id: codec_instance_mgmt.h 73 2012-04-13 13:27:48Z detchart $ */
/*
 * OpenFEC.org AL-FEC Library.
 * (c) Copyright 2009-2012 INRIA - All rights reserved
 * Contact: vincent.roca@inria.fr
 *
 * This software is governed by the CeCILL-C license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-C
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
 */


#ifndef CODEC_INSTANCE_MGMT
#define CODEC_INSTANCE_MGMT
/* AL-FEC extended performance evaluation tool */


/**
 * Creates and initializa a codec instance, either on behalf of a sender or
 * on behalf of a receiver.
 * @param codec_id	
 * @param enc_dec_type	set either to OF_ENCODER or OF_DECODER.	
 * @param k	
 * @param n	
 * @param block		associated block. At a sender, some information may
 *			be read from this block, at a receiver this is the
 *			contrary.	
 * @return		valid pointer to the codec instance created or NULL
 *			in case of error.
 */
of_session_t *
create_and_init_codec_instance (of_codec_id_t	codec_id,
				UINT32		enc_dec_type,
				UINT32		k,
				UINT32		n,
				struct block_cb	*blk);

#endif /* CODEC_INSTANCE_MGMT */
